import traceback

from django.conf import settings
from graphene_django.views import GraphQLView


class OptimizedGraphQLView(GraphQLView):
    def execute_graphql_request(self, *args, **kwargs):
        """
        By default, graphene will eat any exceptions that occur
        Extract any exceptions and echo them to console
        """
        result = super().execute_graphql_request(*args, **kwargs)
        if result and result.errors:
            for error in result.errors:
                try:
                    if hasattr(error, 'original_error'):
                        raise error.original_error
                except Exception as e:
                    if settings.DEBUG:
                        traceback.print_exc()

        return result



import binascii

import graphene
from graphql.language import ast


class Binary(graphene.Scalar):
    """
    BinaryArray is used to convert a Django BinaryField to the string form
    """

    @staticmethod
    def binary_to_string(value):
        return binascii.hexlify(value).decode("utf-8")

    serialize = binary_to_string
    parse_value = binary_to_string

    @classmethod
    def parse_literal(cls, node):
        if isinstance(node, ast.StringValue):
            return cls.binary_to_string(node.value)



import graphene
import graphql_jwt
from django.apps import apps
from graphene import relay
from graphene.utils.str_converters import to_snake_case
from rest_framework.serializers import ModelSerializer

from autographql.models import GraphQLModelBase
from autographql.mutation import DjangoSerializerMutationFieldFactory
from autographql.query import DjangoQueryFactory
from autographql.utils import get_meta
from autographql.types import AutoDjangoObjectType

ALLOWED_ACTIONS = ['retrieve', 'list', 'create', 'update', 'delete']


class RelayDjangoSerializerSchemaFactory(object):
    class Meta:
        fields = None
        node_type = None
        type_name = None
        serializer_class = None
        allowed_actions = None
        retrieve_attribute_name = None
        list_attribute_name = None
        create_attribute_name = None
        update_attribute_name = None
        delete_attribute_name = None

    @classmethod
    def build(cls):
        model = get_meta(cls.Meta, 'model', None)
        fields = get_meta(cls.Meta, 'fields', None)
        node_type = get_meta(cls.Meta, 'node_type', None)
        serializer_class = get_meta(cls.Meta, 'serializer_class', None)

        # Get model class
        if model:
            model_class = model
        elif node_type:
            model_class = node_type._meta.model
        elif serializer_class:
            model_class = serializer_class.Meta.model
        else:
            raise Exception('None of model, node_type, or serializer_class is defined.')

        model_name = model_class.__name__
        model_name_snaked = to_snake_case(model_name)
        type_name = get_meta(cls.Meta, 'type_name', model_name)
        allowed_actions = get_meta(cls.Meta, 'allowed_actions', ALLOWED_ACTIONS)
        b_retrieve_attribute_name = get_meta(cls.Meta, 'retrieve_attribute_name', model_name_snaked)
        b_list_attribute_name = get_meta(cls.Meta, 'list_attribute_name', 'list_' + model_name_snaked)
        b_create_attribute_name = get_meta(cls.Meta, 'create_attribute_name', 'create_' + model_name_snaked)
        b_update_attribute_name = get_meta(cls.Meta, 'update_attribute_name', 'update_' + model_name_snaked)
        b_delete_attribute_name = get_meta(cls.Meta, 'delete_attribute_name', 'delete_' + model_name_snaked)

        # Autogenerate Type
        if node_type:
            Type = node_type
        else:
            type_fields = fields if fields else '__all__'
            Type = type(type_name, (AutoDjangoObjectType,), {
                'Meta': type('Meta', (object,), {
                    'model': model_class,
                    'interfaces': (relay.Node,),
                    'fields': type_fields,
                })
            })

        # Autogenerate SerializerClass
        if serializer_class:
            schema_serializer_class = serializer_class
        else:
            SerializerClass = type(type_name + 'Serializer', (ModelSerializer,), {
                'Meta': type('Meta', (object,), {
                    'model': model_class,
                    'fields': '__all__',
                })
            })
            schema_serializer_class = SerializerClass

        class QueryFactory(DjangoQueryFactory):
            class Meta:
                type = Type
                retrieve = 'retrieve' in allowed_actions
                list = 'list' in allowed_actions
                retrieve_attribute_name = b_retrieve_attribute_name
                list_attribute_name = b_list_attribute_name

        class MutationFieldFactory(DjangoSerializerMutationFieldFactory):
            class Meta:
                type = Type
                serializer_class = schema_serializer_class

        CreateMutation = MutationFieldFactory.get_create_mutation()
        UpdateMutation = MutationFieldFactory.get_update_mutation()
        DeleteMutation = MutationFieldFactory.get_delete_mutation()

        class Mutation(object):
            if 'create' in allowed_actions:
                vars()[b_create_attribute_name] = CreateMutation.Field()
            if 'update' in allowed_actions:
                vars()[b_update_attribute_name] = UpdateMutation.Field()
            if 'delete' in allowed_actions:
                vars()[b_delete_attribute_name] = DeleteMutation.Field()

        return (
            Type,
            QueryFactory.build_query(),
            Mutation
        )


class SchemaGenerator(object):
    """
    Class to automagically create the schema via introspection
    """
    @classmethod
    def get_mutation(cls):
        models = apps.get_models()
        mutation_classes = []
        for model in models:
            if isinstance(model, GraphQLModelBase):
                mutation_classes.append(model._graphql_meta.mutation)
        Mutation = type('Mutation', (*mutation_classes, graphene.ObjectType,), {
            'token_auth': graphql_jwt.relay.ObtainJSONWebToken.Field(),
            'verify_token': graphql_jwt.relay.Verify.Field(),
            'refresh_token': graphql_jwt.relay.Refresh.Field(),
            'revoke_token': graphql_jwt.relay.Revoke.Field(),
        })

        return Mutation

    @classmethod
    def get_query(cls):
        models = apps.get_models()
        query_classes = []
        for model in models:
            if isinstance(model, GraphQLModelBase):
                query_classes.append(model._graphql_meta.query)

        Query = type('Query', (*query_classes, graphene.ObjectType,), {
            'node': relay.Node.Field(),
        })

        return Query

    @classmethod
    def get_schema(cls):
        Query = cls.get_query()
        Mutation = cls.get_mutation()
        schema = graphene.Schema(query=Query, mutation=Mutation)
        return schema


schema = SchemaGenerator.get_schema()



from bridgekeeper import perms
from django.core.exceptions import FieldError
from django.db import models

from autographql.auth.utils import get_model_permission, VIEW


class AuthQuerySet(models.QuerySet):
    def for_user(self, user):
        # Default functionality is identical to all()
        queryset = self.all()

        # Super users can see everything
        if user.is_superuser:
            return queryset

        # Filter queryset based on user permissions
        permission = get_model_permission(self.model, VIEW)
        if permission in perms:
            queryset = perms[permission].filter(user, queryset)

        return queryset


class AuthModelManager(models.Manager):
    def get_queryset(self):
        return AuthQuerySet(self.model, using=self._db)

    def for_user(self, user):
        return self.get_queryset().for_user(user)



import logging
from functools import partial
from inspect import isclass

import graphene
from bridgekeeper import perms
from django.core.exceptions import PermissionDenied
from django.db.models import Q, Field, Transform, Lookup, ForeignObjectRel
from django.db.models.constants import LOOKUP_SEP
from django.db.models.query_utils import RegisterLookupMixin
from graphene import InputField, List, Dynamic
from graphene.types.inputobjecttype import InputObjectTypeOptions
from graphene.types.scalars import ScalarOptions
from graphene_django.registry import get_global_registry
from graphene_django.utils import get_model_fields

from autographql.auth.utils import get_model_permission, VIEW
from autographql.filters.converters import get_input_type_from_lookup
from autographql.filters.fields import LogicalInputField, AND, OR, NOT, LogicalAndInputField, LogicalOrInputField, \
    LogicalNotInputField
from autographql.filters.utils import get_input_type
from autographql.utils import to_pascal_case

logger = logging.getLogger(__name__)


class AuthAutoInputObjectTypeOptions(InputObjectTypeOptions):
    model = None
    input_fields = None


class AutoFilterInputObjectType(graphene.InputObjectType):
    def get_q_lookup(self):
        lookups = self._get_lookups()
        lookup = self._get_q_lookup_helper(lookups)
        return lookup

    def _get_q_lookup_helper(self, lookup, operator=AND):
        """Recursive helper to build the Q lookup"""
        # Base case, we have a tuple
        if isinstance(lookup, tuple):
            args = {lookup[0]: lookup[1]}
            return Q(**args)

        # We have a list, recur for each element and and them together
        if isinstance(lookup, list) and len(lookup) >= 1:
            result = None
            for l in lookup:
                r = self._get_q_lookup_helper(l)
                if not result:
                    result = r
                else:
                    if operator == AND:
                        result = result & r
                    elif operator == OR:
                        result = result | r
            return result

        # We have a dict, use the logical operator stored to process the result
        if isinstance(lookup, dict) and lookup:
            op, records = list(lookup.items())[0]
            if op == NOT:
                l = self._get_q_lookup_helper(records)
                return ~l
            else:
                return self._get_q_lookup_helper(records, op)

    def _is_scalar(self, field):
        t = field.type
        if isinstance(t, List):
            t = t.of_type

        if hasattr(t, '_meta') and isinstance(t._meta, ScalarOptions):
            return True
        return False

    def _get_lookups(self, lookup_path='', **kwargs):
        """Helper function to build lookups"""
        lookups = []
        for field_name, field in self._meta.fields.items():
            attr_value = getattr(self, field_name, None)
            if not attr_value:
                # This filter is not set
                continue

            if lookup_path:
                cur_path = LOOKUP_SEP.join([lookup_path, field_name])
            else:
                cur_path = field_name

            if self._is_scalar(field):
                # Base case, scalar value or list of scalar value
                lookups.append((cur_path, attr_value))
                continue

            if isinstance(attr_value, AutoFilterInputObjectType):
                # Recursive call to get the lookups
                child_lookups = attr_value._get_lookups(cur_path, **kwargs)
                lookups += child_lookups
                continue

            if isinstance(field.type, LogicalInputField):
                operator = field.type.get_operator()
                operator_lookups = {operator: []}
                for child in attr_value:
                    child_lookups = child._get_lookups(lookup_path, **kwargs)
                    operator_lookups[operator] += child_lookups
                lookups.append(operator_lookups)
                continue

        return lookups


class AuthAutoFilterInputObjectType(AutoFilterInputObjectType):
    """
    Input object that only allows users who have permission to view the model to
    use it as a filter.
    """
    @classmethod
    def __init_subclass_with_meta__(cls, input_fields=None, model=None, container=None, _meta=None, **options):
        if not _meta:
            _meta = AuthAutoInputObjectTypeOptions(cls)

        if not model:
            raise RuntimeError('model is required in Meta class for {0}'.format(cls))

        registry = get_global_registry()

        # Add AND, OR, and NOT fields
        def get_self_input_type():
            model_type = registry.get_type_for_model(model)
            return model_type._meta.filter_input_type

        fields = {}
        fields['and'] = InputField.mounted(LogicalAndInputField(
            name='_and',
            of_type=get_self_input_type,
            description='Logical AND is applied to all filters in the contained list.'
        ))
        fields['or'] = InputField.mounted(LogicalOrInputField(
            name='_or',
            of_type=get_self_input_type,
            description='Logical OR is applied to all filters in the contained list.'
        ))
        fields['not'] = InputField.mounted(LogicalNotInputField(
            get_self_input_type,
            name='_not',
            description='Logical NOT is applied to the result of contained the filter.'
        ))

        model_fields = get_model_fields(model)
        model_fields = filter(lambda m: m[0] in input_fields, model_fields)
        for name, field in model_fields:
            n, f = cls._get_filter_input(registry, model, field, name=model.__name__)
            fields[n] = f
        _meta.model = model
        _meta.input_fields = input_fields

        if _meta.fields:
            _meta.fields.update(fields)
        else:
            _meta.fields = fields

        super().__init_subclass_with_meta__(container=None, _meta=_meta, **options)

    @classmethod
    def _get_filter_input(cls, registry, model, node, name, field=None):
        """Recursive helper to generate the input filters"""
        # Base case, dynamic field
        if isinstance(registry.get_converted_field(node), Dynamic):
            return node.name, graphene.InputField(partial(get_input_type, node))

        # Base case, field is of a lookup type
        if isclass(node) and issubclass(node, Lookup):
            dummy = node.__new__(node)
            return get_input_type_from_lookup(dummy, field)

        # field registers transforms
        if isclass(node) and issubclass(node, Transform):
            lookups = node.get_lookups()
            if not lookups:
                dummy = node.__new__(node)
                return get_input_type_from_lookup(dummy, field)

            lookup_fields = {}
            for lookup_name, lookup in lookups.items():
                next_name = name + to_pascal_case(lookup_name)
                n, f = cls._get_filter_input(registry, model, lookup, name=next_name, field=lookup)
                lookup_fields[n] = f

            return node.lookup_name, graphene.InputField(
                type(
                    name + to_pascal_case(node.lookup_name) + 'FilterInput',
                    (AutoFilterInputObjectType,),
                    lookup_fields
                )
            )

        # field registers lookups
        if isinstance(node, RegisterLookupMixin):
            lookup_fields = {}
            field_name = node.attname
            next_name = name + to_pascal_case(field_name)

            lookups = node.get_lookups()
            for lookup in lookups.values():
                n, f = cls._get_filter_input(registry, model, lookup, name=next_name, field=node)
                lookup_fields[n] = f

            return field_name, graphene.InputField(
                type(
                    next_name + 'FilterInput',
                    (AutoFilterInputObjectType,),
                    lookup_fields
                )
            )

        raise RuntimeError('Failed generate filter tree')

    def get_q_lookup(self, user=None):
        lookups = self._get_lookups(user=user)
        lookup = self._get_q_lookup_helper(lookups)
        return lookup

    def _get_lookups(self, lookup_path='', user=None):
        lookups = super()._get_lookups(lookup_path, user=user)
        if not lookups:
            return lookups
        # Permission check
        permission = get_model_permission(self._meta.model, VIEW)
        logger.debug('Checking permission {0} for user [{1}] to use filter [{2}]'.format(permission, user, self.__class__))

        if (permission in perms and perms[permission].is_possible_for(user)) or user.has_perm(permission):
            return lookups
        raise PermissionDenied('User {0} does not have permission to use the {1} filter'.format(str(user), self.__class__))



from django.db import models
from django.utils.translation import ugettext as _

from autographql.models import GraphQLModel


class Region(GraphQLModel):
    region_id = models.PositiveSmallIntegerField(
        _('Region id'),
        db_column='RegionID',
        primary_key=True
    )
    region_description = models.CharField(
        _('Region description'),
        db_column='RegionDescription',
        max_length=50
    )

    class Meta:
        db_table = 'region'


class Territories(GraphQLModel):
    territory_id = models.CharField(
        _('Territory id'),
        db_column='TerritoryID',
        primary_key=True,
        max_length=20
    )
    territory_description = models.CharField(
        _('Territory description'),
        db_column='TerritoryDescription',
        max_length=50
    )
    region_id = models.ForeignKey(
        Region,
        db_column='RegionID',
        on_delete=models.CASCADE
    )

    class Meta:
        db_table = 'territories'


class Employees(GraphQLModel):
    employee_id = models.AutoField(
        _('Employee id'),
        db_column='EmployeeID',
        primary_key=True,
    )
    last_name = models.CharField(
        _('Last name'),
        db_column='LastName',
        max_length=20,
        db_index=True
    )
    first_name = models.CharField(
        _('First name'),
        db_column='FirstName',
        max_length=10
    )
    title = models.CharField(
        _('title'),
        db_column='Title',
        max_length=30,
        blank=True,
        null=True
    )
    title_of_courtesy = models.CharField(
        _('Title of Courtesy'),
        db_column='TitleOfCourtesy',
        max_length=25,
        blank=True,
        null=True
    )
    birth_date = models.DateTimeField(
        _('Birth date'),
        db_column='BirthDate',
        blank=True,
        null=True
    )
    hire_date = models.DateTimeField(
        _('Hire date'),
        db_column='HireDate',
        blank=True,
        null=True
    )
    address = models.CharField(
        _('address'),
        db_column='Address',
        max_length=60,
        blank=True,
        null=True
    )
    city = models.CharField(
        _('City'),
        db_column='City',
        max_length=15,
        blank=True,
        null=True
    )
    region = models.CharField(
        _('Region'),
        db_column='Region',
        max_length=15,
        blank=True,
        null=True
    )
    postal_code = models.CharField(
        _('Postal code'),
        db_column='PostalCode',
        max_length=10,
        blank=True,
        db_index=True
    )
    country = models.CharField(
        _('Country'),
        db_column='Country',
        max_length=15,
        blank=True,
        null=True
    )
    home_phone = models.CharField(
        _('Home phone'),
        db_column='HomePhone',
        max_length=24,
        blank=True,
        null=True
    )
    extension = models.CharField(
        _('Extension'),
        db_column='Extension',
        max_length=4,
        blank=True,
        null=True
    )
    photo = models.BinaryField(
        _('Photo'),
        db_column='Photo',
        blank=True,
        null=True
    )
    notes = models.TextField(
        _('Notes'),
        db_column='Notes',
        blank=True,
        null=True
    )
    reports_to = models.ForeignKey(
        'self',
        null=True,
        blank=True,
        db_column='ReportsTo',
        on_delete=models.PROTECT,
    )
    photo_path = models.CharField(
        _('Photo path'),
        db_column='PhotoPath',
        max_length=255,
        blank=True,
        null=True
    )
    territories = models.ManyToManyField(
        Territories,
        verbose_name=_('Territories'),
        db_table='EmployeeTerritories',
        blank=True,
    )

    class Meta:
        db_table = 'employees'


class Shippers(GraphQLModel):
    shipper_id = models.AutoField(
        _('Shipper ID'),
        db_column='ShipperID',
        primary_key=True
    )
    company_name = models.CharField(
        _('Company name'),
        db_column='CompanyName',
        max_length=40
    )
    phone = models.CharField(
        _('Phone'),
        db_column='Phone',
        max_length=24,
        blank=True,
        null=True
    )

    class Meta:
        db_table = 'shippers'


class CustomerDemographics(GraphQLModel):
    customer_type_id = models.CharField(
        _('Customer typeID'),
        db_column='CustomerTypeID',
        primary_key=True,
        max_length=5
    )
    customer_desc = models.TextField(
        _('Customer description'),
        db_column='CustomerDesc',
        blank=True,
        null=True
    )

    class Meta:
        db_table = 'customerdemographics'


class Customers(GraphQLModel):
    customer_id = models.CharField(
        _('Customer ID'),
        db_column='CustomerID',
        primary_key=True,
        max_length=5
    )
    company_name = models.CharField(
        _('Company name'),
        db_column='CompanyName',
        max_length=40
    )
    contact_name = models.CharField(
        _('Contact name'),
        db_column='ContactName',
        max_length=30,
        blank=True,
        null=True
    )
    contact_title = models.CharField(
        _('Contact title'),
        db_column='ContactTitle',
        max_length=30,
        blank=True,
        null=True
    )
    address = models.CharField(
        _('Address'),
        db_column='Address',
        max_length=60,
        blank=True,
        null=True
    )
    city = models.CharField(
        _('City'),
        db_column='City',
        max_length=15,
        blank=True,
        null=True
    )
    region = models.CharField(
        _('Region'),
        db_column='Region',
        max_length=15,
        blank=True,
        null=True
    )
    postal_code = models.CharField(
        _('Postal code'),
        db_column='PostalCode',
        max_length=10,
        blank=True,
        null=True
    )
    country = models.CharField(
        _('Country'),
        db_column='Country',
        max_length=15,
        blank=True,
        null=True
    )
    phone = models.CharField(
        _('Phone'),
        db_column='Phone',
        max_length=24,
        blank=True,
        null=True
    )
    fax = models.CharField(
        _('Fax'),
        db_column='Fax',
        max_length=24,
        blank=True,
        null=True
    )
    customer_customer_demo = models.ManyToManyField(
        CustomerDemographics,
        verbose_name=_('Customer customer demo'),
        db_table='CustomerCustomerDemo',
        blank=True,
    )

    class Meta:
        db_table = 'customers'


class Categories(GraphQLModel):
    category_id = models.AutoField(
        _('Category ID'),
        db_column='CategoryID',
        primary_key=True
    )
    category_name = models.CharField(
        _('Category name'),
        db_column='CategoryName',
        max_length=15,
        db_index=True
    )
    description = models.TextField(
        _('Description'),
        db_column='Description',
        blank=True,
        null=True
    )
    picture = models.BinaryField(
        _('Picture'),
        db_column='Picture',
        blank=True,
        null=True
    )

    class Meta:
        db_table = 'categories'


class Suppliers(GraphQLModel):
    supplier_id = models.AutoField(
        _('Supplier ID'),
        db_column='SupplierID',
        primary_key=True
    )
    company_name = models.CharField(
        _('Company name'),
        db_column='CompanyName',
        max_length=40
    )
    contact_name = models.CharField(
        _('Contact name'),
        db_column='ContactName',
        max_length=30,
        blank=True,
        null=True
    )
    contact_title = models.CharField(
        _('Contact title'),
        db_column='ContactTitle',
        max_length=30,
        blank=True,
        null=True
    )
    address = models.CharField(
        _('Address'),
        db_column='Address',
        max_length=60,
        blank=True,
        null=True
    )
    city = models.CharField(
        _('City'),
        db_column='City',
        max_length=15,
        blank=True,
        null=True
    )
    region = models.CharField(
        _('Region'),
        db_column='Region',
        max_length=15,
        blank=True,
        null=True
    )
    postal_code = models.CharField(
        _('Postal code'),
        db_column='PostalCode',
        max_length=10,
        blank=True,
        null=True
    )
    country = models.CharField(
        _('Country'),
        db_column='Country',
        max_length=15,
        blank=True,
        null=True
    )
    phone = models.CharField(
        _('Phone'),
        db_column='Phone',
        max_length=24,
        blank=True,
        null=True
    )
    fax = models.CharField(
        _('Fax'),
        db_column='Fax',
        max_length=24,
        blank=True,
        null=True
    )
    homepage = models.TextField(
        _('HomePage'),
        db_column='HomePage',
        blank=True,
        null=True
    )

    class Meta:
        db_table = 'suppliers'


class Products(GraphQLModel):
    product_id = models.AutoField(
        _('Product ID'),
        db_column='ProductID',
        primary_key=True
    )
    product_name = models.CharField(
        _('Product name'),
        db_column='ProductName',
        max_length=40,
        db_index=True,
    )
    supplier_id = models.ForeignKey(
        Suppliers,
        db_column='SupplierID',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        db_index=True,
    )
    category_id = models.ForeignKey(
        Categories,
        db_column='CategoryID',
        db_index=True,
        blank=True,
        null=True,
        on_delete=models.CASCADE
    )
    quantity_per_unit = models.CharField(
        _('Quantity per Unit'),
        db_column='QuantityPerUnit',
        max_length=20,
        blank=True,
        null=True
    )
    unit_price = models.DecimalField(
        _('Unit price'),
        db_column='UnitPrice',
        blank=True,
        null=True,
        max_digits=19,
        decimal_places=4
    )
    units_in_stock = models.SmallIntegerField(
        _('Units in Stock'),
        db_column='UnitsInStock',
        blank=True,
        null=True
    )
    units_on_order = models.SmallIntegerField(
        _('Units on Order'),
        db_column='UnitsOnOrder',
        blank=True,
        null=True
    )
    reorder_level = models.SmallIntegerField(
        _('Reorder level'),
        db_column='ReorderLevel',
        blank=True,
        null=True
    )
    discontinued = models.IntegerField(
        _('Discontinued'),
        db_column='Discontinued'
    )

    class Meta:
        db_table = 'products'


class Orders(GraphQLModel):
    order_id = models.AutoField(
        _('Order ID'),
        db_column='OrderID',
        primary_key=True
    )
    customer = models.ForeignKey(
        Customers,
        blank=True,
        null=True,
        db_column='CustomerID',
        on_delete=models.CASCADE,
        db_index=True
    )
    employee = models.ForeignKey(
        Employees,
        blank=True,
        null=True,
        db_column='EmployeeID',
        on_delete=models.CASCADE,
        db_index=True
    )
    order_date = models.DateField(
        _('Order date'),
        db_column='OrderDate',
        blank=True,
        null=True,
        db_index=True
    )
    required_date = models.DateField(
        _('Required_date'),
        db_column='RequiredDate',
        blank=True,
        null=True
    )
    shipped_date = models.DateField(
        _('Shipped date'),
        db_column='ShippedDate',
        blank=True,
        null=True,
        db_index=True
    )
    ship_via = models.ForeignKey(
        Shippers,
        blank=True,
        null=True,
        db_column='ShipVia',
        on_delete=models.CASCADE,
        db_index=True
    )
    freight = models.DecimalField(
        _('Freight'),
        db_column='Freight',
        blank=True,
        null=True,
        max_digits=19,
        decimal_places=4
    )
    ship_name = models.CharField(
        _('Ship name'),
        db_column='ShipName',
        max_length=40,
        blank=True,
        null=True
    )
    ship_address = models.CharField(
        _('Ship address'),
        db_column='ShipAddress',
        max_length=60,
        blank=True,
        null=True
    )
    ship_city = models.CharField(
        _('Ship city'),
        db_column='ShipCity',
        max_length=15,
        blank=True,
        null=True
    )
    ship_region = models.CharField(
        _('Ship region'),
        db_column='ShipRegion',
        max_length=15,
        blank=True,
        null=True
    )
    ship_postal_code = models.CharField(
        _('Ship postal code'),
        db_column='ShipPostalCode',
        max_length=10,
        blank=True,
        null=True,
        db_index=True
    )
    ship_country = models.CharField(
        _('Shipped country'),
        db_column='ShipCountry',
        max_length=15,
        blank=True,
        null=True
    )
    order_details = models.ManyToManyField(
        Products,
        verbose_name=_('Products'),
        blank=True,
        through='OrderDetails'
    )

    class Meta:
        db_table = 'orders'


class OrderDetails(GraphQLModel):
    order_id = models.ForeignKey(
        Orders,
        db_column='OrderID',
        on_delete=models.CASCADE
    )
    product_id = models.ForeignKey(
        Products,
        db_column='ProductID',
        on_delete=models.CASCADE
    )
    unit_price = models.DecimalField(
        _('Unit price'),
        db_column='UnitPrice',
        max_digits=19,
        decimal_places=4
    )
    quantity = models.SmallIntegerField(
        _('Quantity'),
        db_column='Quantity'
    )
    discount = models.FloatField(
        _('Discount'),
        db_column='Discount'
    )

    class Meta:
        db_table = 'order_details'



from graphene_django.registry import Registry as _Registry


class Registry(_Registry):
    def __init__(self):
        super(Registry, self).__init__()
        self.serializer_create_input_registry = {}
        self.serializer_update_input_registry = {}

    def register_serializer_create_input_field(self, field, converted):
        self.serializer_create_input_registry[field] = converted

    def register_serializer_update_input_field(self, field, converted):
        self.serializer_update_input_registry[field] = converted

    def get_serializer_create_input_field(self, field):
        return self.serializer_create_input_registry.get(field)

    def get_serializer_update_input_field(self, field):
        return self.serializer_update_input_registry.get(field)



from graphene import relay
from graphene.utils.str_converters import to_snake_case

from autographql.fields import OptimizedDjangoConnectionField


class DjangoQueryFactory(object):
    class Meta:
        type = None
        retrieve_attribute_name = None
        list_attribute_name = None
        retrieve = True
        list = True
        resolve_decorators = []

    @classmethod
    def build_query(cls):
        if cls.Meta.type is None:
            # type must be defined
            raise Exception('type meta must be defined!')

        if hasattr(cls.Meta, 'retrieve') and cls.Meta.retrieve is False:
            # No retrieve
            retrieve_attribute_name = None
        elif hasattr(cls.Meta, 'retrieve_attribute_name') and cls.Meta.retrieve_attribute_name is not None:
            retrieve_attribute_name = cls.Meta.retrieve_attribute_name
        else:
            # Default to snake case model name
            retrieve_attribute_name = to_snake_case(cls.Meta.type._meta.model.__name__)

        if hasattr(cls.Meta, 'list') and cls.Meta.list is False:
            # No list
            list_attribute_name = None
        elif hasattr(cls.Meta, 'list_attribute_name') and cls.Meta.list_attribute_name is not None:
            list_attribute_name = cls.Meta.list_attribute_name
        else:
            # Default to list_ + snake cased model name
            list_attribute_name = 'list_' + to_snake_case(cls.Meta.type._meta.model.__name__)

        class Query(object):
            if retrieve_attribute_name is not None:
                vars()[retrieve_attribute_name] = relay.Node.Field(cls.Meta.type)

            if list_attribute_name is not None:
                vars()[list_attribute_name] = OptimizedDjangoConnectionField(cls.Meta.type)

        return Query



from django.utils.functional import cached_property


class GraphQLOptions:
    def __init__(self, model, graphqlmeta):
        self.model = model
        self.meta = graphqlmeta

    @cached_property
    def schema_factory(self):
        from autographql.schema import RelayDjangoSerializerSchemaFactory
        return type(self.model.__name__ + 'SchemaFactory', (RelayDjangoSerializerSchemaFactory,), {
            'Meta': type('Meta', (object, ), {
                'model': self.model,
                'fields': getattr(self.meta, 'fields', None),
            })
        })

    @cached_property
    def schema_factory_output(self):
        return self.schema_factory.build()

    @property
    def node_type(self):
        return self.schema_factory_output[0]

    @property
    def query(self):
        return self.schema_factory_output[1]

    @property
    def mutation(self):
        return self.schema_factory_output[2]



from bridgekeeper.rules import Attribute, Is, Relation as _Relation, ManyRelation as _ManyRelation, UNIVERSAL, EMPTY


def is_rule(f):
    return Is(f)


def attribute_rule(attr):
    def wrapper(matches, repr_string=None):
        return Attribute(attr, matches)
    return wrapper


class Relation(_Relation):
    def check(self, user, instance=None):
        if not hasattr(instance, self.attr):
            return False
        return super(Relation, self).check(user, instance)


class ManyRelation(_ManyRelation):
    def check(self, user, instance=None):
        if instance is None:
            return self.rule.check(user, None)
        attr = instance.__class__._meta.get_field(self.query_attr,).get_accessor_name()
        related_manager = getattr(instance, attr)

        qs = related_manager.get_queryset()
        if qs._prefetch_done:
            # Data already cached, we should use it
            for value in qs:
                if self.rule.check(user, value):
                    return True
            return False

        related_q = self.rule.query(user)
        if related_q is UNIVERSAL or related_q is EMPTY:
            return related_q
        return related_manager.filter(related_q).exists()



import graphene
from django.db import models
from graphene import Dynamic
from graphene_django import DjangoListField
from graphene_django.registry import get_global_registry
from graphene_django.rest_framework.serializer_converter import get_graphene_type_from_serializer_field
from rest_framework import serializers
from rest_framework.fields import HiddenField
from graphene_django.converter import convert_django_field, get_django_field_description

from autographql.base_types import Binary
from autographql.fields import AutoDjangoConnectionField


@get_graphene_type_from_serializer_field.register(serializers.RelatedField)
def convert_related_field_to_global_id(field):
    return graphene.GlobalID


@get_graphene_type_from_serializer_field.register(serializers.IntegerField)
def convert_serializer_field_to_int(field):
    if field.field_name == 'id':
        return graphene.GlobalID
    return graphene.Int


@get_graphene_type_from_serializer_field.register(serializers.ChoiceField)
def convert_serializer_field_to_enum(field):
    """This was added in graphene-django 2.6.0, causes multiple mutations made from a serializer to break"""
    return graphene.String


def convert_serializer_field(field, method, is_input=True, force_required=None):
    """
    Converts a django rest frameworks field to a graphql field
    and marks the field as required if we are creating an input type
    and the field itself is required
    """

    graphql_type = get_graphene_type_from_serializer_field(field)

    args = []
    if force_required is not None:
        required = force_required
    else:
        required = is_input and field.required
    kwargs = {"description": field.help_text, "required": required}

    # if it is a tuple or a list it means that we are returning
    # the graphql type and the child type
    if isinstance(graphql_type, (list, tuple)):
        kwargs["of_type"] = graphql_type[1]
        graphql_type = graphql_type[0]

    if isinstance(field, serializers.ModelSerializer):
        if is_input:
            graphql_type = convert_serializer_to_input_type(field.__class__, method)
        else:
            global_registry = get_global_registry()
            field_model = field.Meta.model
            args = [global_registry.get_type_for_model(field_model)]
    elif isinstance(field, serializers.ListSerializer):
        field = field.child
        if is_input:
            kwargs["of_type"] = convert_serializer_to_input_type(field.__class__, method)
        else:
            del kwargs["of_type"]
            global_registry = get_global_registry()
            field_model = field.Meta.model
            args = [global_registry.get_type_for_model(field_model)]

    return graphql_type(*args, **kwargs)


def get_input_fields_from_serializer(serializer, method):
    items = {}
    for name, field in serializer.fields.items():
        # Always skip hidden fields
        if isinstance(field, HiddenField):
            continue

        # Skip read only fields except for the id field
        if field.read_only and name != 'id':
            continue

        if method == 'create':
            # id field should be skipped on create
            if name == 'id':
                continue

            items[name] = convert_serializer_field(field, method)

        elif method == 'update':
            # all fields should be not required except id
            if name == 'id':
                items[name] = convert_serializer_field(field, method, force_required=True)
                continue
            items[name] = convert_serializer_field(field, method, force_required=False)

        else:
            raise ValueError('method must be create or update')

    return items


def convert_serializer_to_input_type(serializer_class, method):
    """
    Modified converter to cache the serializer input type and reuse it
    If we don't cache it, graphene_django will recreate the input type for nested serializers
    each time which results in the same input type being declared twice
    """
    global_registry = get_global_registry()
    if method == 'create':
        get_field = global_registry.get_serializer_create_input_field
        register_field = global_registry.register_serializer_create_input_field
    elif method == 'update':
        get_field = global_registry.get_serializer_update_input_field
        register_field = global_registry.register_serializer_update_input_field
    else:
        raise ValueError('method argument must be either create or update')

    input_type = get_field(serializer_class)
    if input_type:
        return input_type

    serializer = serializer_class()

    items = get_input_fields_from_serializer(serializer, method)

    input_type = type(
        "{}{}Input".format(serializer.__class__.__name__, method.capitalize()),
        (graphene.InputObjectType,),
        items,
    )
    register_field(serializer_class, input_type)

    return input_type


@convert_django_field.register(models.ManyToManyField)
@convert_django_field.register(models.ManyToManyRel)
@convert_django_field.register(models.ManyToOneRel)
def convert_field_to_list_or_connection(field, registry=None):
    model = field.related_model

    def dynamic_type():
        _type = registry.get_type_for_model(model)
        if not _type:
            return

        if isinstance(field, models.ManyToManyField):
            description = get_django_field_description(field)
        else:
            description = get_django_field_description(field.field)

        # If there is a connection, we should transform the field
        # into a DjangoConnectionField
        if _type._meta.connection:
            return AutoDjangoConnectionField(_type, required=True, description=description)

        return DjangoListField(
            _type,
            required=True,  # A Set is always returned, never None.
            description=description,
        )

    return Dynamic(dynamic_type)


@convert_django_field.register(models.BinaryField)
def convert_binary_to_string(field, registry=None):
    return Binary(
        description=get_django_field_description(field),
        required=not field.null,
    )



import logging

from bridgekeeper import perms
from django.core.exceptions import PermissionDenied
from django.db.models import Model
from graphene import Connection

from autographql.auth.utils import VIEW, get_model_permission
from autographql.auth.constants import PERMISSION_DENIED_MESSAGE

logger = logging.getLogger(__name__)


class AuthorizationMiddleware(object):
    def resolve(self, next, root, info, **args):
        instance = next(root, info, **args)

        if info.context.user.is_superuser:
            # Superusers have all permissions
            return instance

        value = instance
        user = info.context.user

        # If model, check user has permission to access this model instance
        if isinstance(value, Model):
            # Check access permissions
            permission = get_model_permission(value._meta.model, VIEW)
            logger.debug('Checking permission {0} for user [{1}] to access instance'.format(permission, user))
            if not user.has_perm(permission, value):
                logger.debug('Permission denied for user [{0}] to instance'.format(user))
                raise PermissionDenied(PERMISSION_DENIED_MESSAGE)

        # Connection field, check user has permissions to list this model's instances
        elif isinstance(value, Connection):
            # Make sure it has a model meta on the type. If it doesn't probably not a django connection
            if hasattr(value._meta.node._meta, 'model'):
                model = value._meta.node._meta.model
                permission = get_model_permission(model, VIEW)
                logger.debug('Checking permission {0} for user [{1}] to list [{2}]'.format(permission, user, value))

                # Check bridgekeeper first for possible permissions
                if permission in perms and perms[permission].is_possible_for(user):
                    pass
                # Check against django auth chain
                elif user.has_perm(permission):
                    pass
                else:
                    logger.debug('Permission denied for user [{0}] to list [{1}]'.format(user, value))
                    raise PermissionDenied(PERMISSION_DENIED_MESSAGE)

        return instance



VIEW = 'view'
CREATE = 'create',
UPDATE = 'update',
DELETE = 'delete'

permission_map = {
    VIEW: '%(app_label)s.view_%(model_name)s',
    CREATE: '%(app_label)s.add_%(model_name)s',
    UPDATE: '%(app_label)s.change_%(model_name)s',
    DELETE: '%(app_label)s.delete_%(model_name)s',
}


def get_model_permission(model, type):
    kwargs = {
        'app_label': model._meta.app_label,
        'model_name': model._meta.model_name
    }
    return permission_map[type] % kwargs



from bridgekeeper import perms
from bridgekeeper.rules import BinaryCompositeRule, Attribute, Relation, ManyRelation
from django.core.exceptions import FieldDoesNotExist
from django.db.models import Prefetch
from django.db.models.fields.related import ForeignKey

from autographql.auth.utils import get_model_permission, VIEW


class AuthQueryOptimizer(object):
    def __init__(self, info):
        self.root_info = info
        self.user = info.context.user

    def optimize(self, queryset):
        permission = get_model_permission(queryset.model, VIEW)
        if permission not in perms:
            # Nothing to optimize
            return queryset

        rule = perms[permission]

        queryset = self.optimize_queryset(queryset, rule)
        return queryset

    def optimize_queryset(self, queryset, rule):
        store = AuthQueryOptimizerStore()
        queryset = self.optimize_queryset_rule(queryset, store, rule)
        return store.optimize(queryset)

    def get_queryset(self, model):
        return model.objects.for_user(self.user)

    def optimize_queryset_rule(self, queryset, store, rule):
        """Recursive helper to process rule tree"""
        # BinaryCompositeRule
        if isinstance(rule, BinaryCompositeRule):
            # Left and right, recursive call
            queryset = self.optimize_queryset_rule(queryset, store, rule.left)
            queryset = self.optimize_queryset_rule(queryset, store, rule.right)

        # Relation rule
        elif isinstance(rule, Relation):
            model_field = self.get_model_field_from_name(queryset.model, rule.attr)
            queryset = queryset.prefetch_related(
                Prefetch(
                    rule.attr,
                    queryset=self.optimize_queryset(
                        self.get_queryset(model_field.related_model),
                        rule.rule,
                    )
                )
            )

        # ManyRelation rule
        elif isinstance(rule, ManyRelation):
            attr = queryset.model.__class__._meta.get_field(rule.query_attr,).get_accessor_name()
            model_field = self.get_model_field_from_name(queryset.model, attr)
            queryset = queryset.prefetch_related(
                Prefetch(
                    attr,
                    queryset=self.optimize_queryset(
                        self.get_queryset(model_field.related_model),
                        rule.rule,
                    )
                )
            )

        # Attribute rule
        elif isinstance(rule, Attribute):
            model_field = self.get_model_field_from_name(queryset.model, rule.attr)
            store.optimize_field(model_field, rule.attr)

        return queryset
    
    def get_model_field_from_name(self, model, name):
        try:
            return model._meta.get_field(name)
        except FieldDoesNotExist:
            descriptor = model.__dict__.get(name)
            if not descriptor:
                return None
            return getattr(descriptor, 'rel', None) \
                   or getattr(descriptor, 'related', None)


class AuthQueryOptimizerStore(object):
    def __init__(self):
        self.select_list = []
        self.prefetch_list = []
        self.only_list = []

    def optimize(self, queryset):
        # Add the selects to the queryset
        if len(self.select_list) > 0:
            queryset = queryset.select_related(*self.select_list)

        # Add the prefetches to the queryest
        if len(self.prefetch_list) > 0:
            queryset = queryset.prefetch_related(*self.prefetch_list)

        # Add the onlys to the queryset
        # .only overwrites old onlys so we need to pull the existing onlys from the queryset
        if len(self.only_list) > 0:
            only_list = self.only_list
            loaded_field_names = queryset.query.get_loaded_field_names()
            if loaded_field_names:
                only_list = list(loaded_field_names[queryset.model]) + only_list
                queryset = queryset.only(*only_list)

        return queryset
    
    def optimize_field(self, model_field, field):
        if self._is_foreign_key_id(model_field, field):
            self.only(field)
            return True
        if model_field.many_to_one or model_field.one_to_one:
            self.select_related(field)
            return True
        if model_field.one_to_many or model_field.many_to_many:
            self.prefetch_related(field)
            return True
        return False

    def only(self, field):
        if field not in self.only_list:
            self.only_list.append(field)

    def select_related(self, field):
        if field not in self.select_list:
            self.select_list.append(field)

    def prefetch_related(self, field):
        if field not in self.prefetch_list:
            self.prefetch_list.append(field)

    def _is_foreign_key_id(self, model_field, name):
        return (
            isinstance(model_field, ForeignKey) and
            model_field.name != name and
            model_field.get_attname() == name
        )



from functools import partial

from django.db.models import Model, QuerySet
from graphene import Field, Dynamic
from graphene_django import DjangoObjectType, DjangoConnectionField

from autographql.types import OrderByType


class AutoDjangoConnectionField(DjangoConnectionField):
    """
    Customized DjangoConnectionField to add order and filter fields
    """
    def __init__(self, type_, *args, **kwargs):
        kwargs.setdefault('orderBy', OrderByType())
        kwargs.setdefault('where', Dynamic(lambda: type_._meta.filter_input_type()))
        super().__init__(type_, *args, **kwargs)


class OptimizedDjangoConnectionField(AutoDjangoConnectionField):
    """
    Optimized connection field, applies all optimizations and filters to the queryset
    Should only be used at the root level queries
    """
    @classmethod
    def resolve_queryset(cls, connection, queryset, info, args):
        qs = super().resolve_queryset(connection, queryset, info, args)
        qs = connection._meta.node.optimize_queryset(qs, info, args)
        return qs


class OptimizedField(Field):
    def resolve_optimized_field(self, parent_resolver, root, info, **args):
        if not issubclass(self.type, DjangoObjectType):
            return parent_resolver(root, info, **args)

        if hasattr(parent_resolver, 'args'):
            attr_name = parent_resolver.args[0]
            value = getattr(root, attr_name)
        else:
            value = parent_resolver(root, info, **args)

        if isinstance(value, Model):
            # Model instance, need to use it to get new queryset
            return self.type.get_optimized_queryset(info, args).get(pk=value.pk)
        elif isinstance(value, QuerySet):
            # Received queryset, optimize it
            return self.type.optimize_queryset(value, info, args).get()

        return value

    def get_resolver(self, parent_resolver):
        if self.resolver:
            return self.resolver

        return partial(self.resolve_optimized_field, parent_resolver)



from django.apps import AppConfig


class AutographqlConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'autographql'

    def ready(self):
        import autographql.converters
        import autographql.filters.converters
        import autographql.monkeypatch



from enum import Enum
from functools import singledispatch

import graphene
from django.core.exceptions import ImproperlyConfigured
from django.db.models import lookups
from django.db.models.functions import ExtractYear, ExtractMonth, ExtractDay, ExtractWeekDay, ExtractIsoWeekDay, \
    ExtractWeek, ExtractIsoYear, ExtractQuarter, ExtractHour, ExtractMinute, ExtractSecond, TruncDate, TruncTime
from graphene import Int
from graphene_django.converter import convert_django_field, convert_field_to_boolean, convert_field_to_string, \
    get_django_field_description, convert_field_to_int, convert_date_to_string, convert_time_to_string


def get_lookup_name(lookup):
    return lookup.lookup_name


def convert_lookup(converter, lookup, field):
    converted = converter(field)
    cls = converted.__class__
    name = get_lookup_name(lookup)
    return name, cls(
        description=get_django_field_description(field),
        required=False,
    )


@singledispatch
def get_input_type_from_lookup(lookup, field):
    raise ImproperlyConfigured(
        "Don't know how to convert the lookup %s (%s) "
        "to Graphene type" % (lookup, lookup.__class__)
    )


@get_input_type_from_lookup.register(lookups.In)
def convert_lookup_to_list_field_type(lookup, field):
    return get_lookup_name(lookup), graphene.List(
        of_type=convert_django_field(field).__class__,
        description=get_django_field_description(field),
        required=False,
    )


@get_input_type_from_lookup.register(TruncDate)
def convert_lookup_to_date(lookup, field):
    return convert_lookup(convert_date_to_string, lookup, field)


@get_input_type_from_lookup.register(TruncTime)
def convert_lookup_to_date(lookup, field):
    return convert_lookup(convert_time_to_string, lookup, field)


@get_input_type_from_lookup.register(lookups.YearExact)
@get_input_type_from_lookup.register(lookups.YearGt)
@get_input_type_from_lookup.register(lookups.YearGte)
@get_input_type_from_lookup.register(lookups.YearLt)
@get_input_type_from_lookup.register(lookups.YearLte)
@get_input_type_from_lookup.register(ExtractMonth)
@get_input_type_from_lookup.register(ExtractDay)
@get_input_type_from_lookup.register(ExtractWeekDay)
@get_input_type_from_lookup.register(ExtractIsoWeekDay)
@get_input_type_from_lookup.register(ExtractWeek)
@get_input_type_from_lookup.register(ExtractQuarter)
@get_input_type_from_lookup.register(ExtractHour)
@get_input_type_from_lookup.register(ExtractMinute)
@get_input_type_from_lookup.register(ExtractSecond)
def convert_lookup_to_int(lookup, field):
    return get_lookup_name(lookup), Int(required=False)


# def convert_lookup_to_week_day(lookup, field):
#     return get_lookup_name(lookup), Enum(
#         'WeekDay', [
#             ('Sunday', 1),
#             ('Monday', 2),
#             ('Tuesday', 3),
#             ('Wednesday', 4),
#             ('Thursday', 5),
#             ('Friday', 6),
#             ('Saturday', 7),
#         ],
#     )


@get_input_type_from_lookup.register(lookups.IsNull)
def convert_lookup_to_boolean(lookup, field):
    return convert_lookup(convert_field_to_boolean, lookup, field)


@get_input_type_from_lookup.register(lookups.Regex)
@get_input_type_from_lookup.register(lookups.IRegex)
def convert_lookup_to_string(lookup, field):
    return convert_lookup(convert_field_to_string, lookup, field)


@get_input_type_from_lookup.register(lookups.Exact)
@get_input_type_from_lookup.register(lookups.IExact)
@get_input_type_from_lookup.register(lookups.GreaterThan)
@get_input_type_from_lookup.register(lookups.GreaterThanOrEqual)
@get_input_type_from_lookup.register(lookups.IntegerGreaterThanOrEqual)
@get_input_type_from_lookup.register(lookups.IntegerLessThan)
@get_input_type_from_lookup.register(lookups.LessThan)
@get_input_type_from_lookup.register(lookups.LessThanOrEqual)
@get_input_type_from_lookup.register(lookups.Contains)
@get_input_type_from_lookup.register(lookups.IContains)
@get_input_type_from_lookup.register(lookups.StartsWith)
@get_input_type_from_lookup.register(lookups.IStartsWith)
@get_input_type_from_lookup.register(lookups.EndsWith)
@get_input_type_from_lookup.register(lookups.IEndsWith)
@get_input_type_from_lookup.register(lookups.Range)
def convert_lookup_to_field_type(lookup, field):
    return convert_lookup(convert_django_field, lookup, field)



from django.db.models import Prefetch


def remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix):]
    return text


def merge_querysets(queryset_list):
    new_qs = None
    for qs in queryset_list:
        if new_qs is None:
            new_qs = qs
        else:
            new_qs = new_qs & qs

    return new_qs


def combine_querysets(queryset_list):
    """Recursive helper. Combines the querysets in queryset_list and recursively
        deduplicates and applies any prefetches"""
    merged_queryset = merge_querysets(queryset_list)

    # Combine duplicate prefetch types
    prefetch_list = []
    prefetch_map = {}

    for queryset in queryset_list:
        for prefetch in queryset._prefetch_related_lookups:
            if isinstance(prefetch, str):
                # String prefetch, just add it to the list
                prefetch_list.append(prefetch)
                continue

            through = prefetch.prefetch_through
            # If through is in prefetch_list as a string, remove it
            if through in prefetch_list:
                prefetch_list = list(filter(lambda p: p != through, prefetch_list))

            if through not in prefetch_map:
                prefetch_map[through] = []

            prefetch_map[through].append(prefetch.queryset)

    if not prefetch_map.items():
        # Base case, no prefetches, just returned the combined queryset
        return merged_queryset

    for through, prefetch_querysets in prefetch_map.items():
        # Has prefetches, recursive call
        prefetch_qs = combine_querysets(prefetch_querysets)
        prefetch_list.append(Prefetch(through, queryset=prefetch_qs))

    merged_queryset = merged_queryset.prefetch_related(None)
    merged_queryset = merged_queryset.prefetch_related(*prefetch_list)

    return merged_queryset



from django.db import models
from django.db.models.base import ModelBase
from autographql.managers import AuthModelManager
from autographql.options import GraphQLOptions


class GraphQLModelBase(ModelBase):
    def __new__(cls, name, bases, attrs, **kwargs):
        model = super().__new__(cls, name, bases, attrs, **kwargs)
        if model._meta.abstract:
            return model

        attr_graphql_meta = attrs.pop('GraphQLMeta', None)
        model.add_to_class('_graphql_meta', GraphQLOptions(
            model,
            attr_graphql_meta,
        ))

        return model


class GraphQLModel(models.Model, metaclass=GraphQLModelBase):
    """
    Abstract base model for all models
    """
    _graphql_meta: GraphQLOptions = None
    # Default to a custom model manager that has a for_user field
    objects = AuthModelManager()

    class Meta:
        abstract = True



import graphene
from graphql.language.ast import VariableNode, ArgumentNode, ListValueNode, ValueNode, ObjectValueNode

from autographql import types


def get_meta(meta, attribute, default):
    if hasattr(meta, attribute) and getattr(meta, attribute) is not None:
        return getattr(meta, attribute)
    else:
        return default


def get_arg_value(info, value):
    """Helper method to replace variables with values"""
    if isinstance(value, VariableNode):
        try:
            return info.variable_values[value.name.value]
        except KeyError:
            return None
    else:
        return value


def flatten_args(info, args):
    results = {}
    for arg in args:
        value = flatten_arg(info, arg)
        results.update(value)

    return results


def flatten_arg(info, arg):
    """Recursive helper function to parse args"""
    if isinstance(arg, ArgumentNode):
        return {arg.name.value: flatten_arg(info, arg.value)}
    elif isinstance(arg, VariableNode):
        try:
            return flatten_arg(info, info.variable_values[arg.name.value])
        except KeyError:
            # Variable not set
            return None
    elif isinstance(arg, list):
        values = []
        for v in arg:
            values.append(flatten_arg(info, v))
        return values
    elif isinstance(arg, graphene.InputObjectType):
        value = dict(arg)
        for k in value.keys():
            value[k] = flatten_arg(info, value[k])
        return value
    elif isinstance(arg, ObjectValueNode):
        value = {}
        for field in arg.fields:
            value[field.name.value] = flatten_arg(info, field.value)
        return value
    elif isinstance(arg, ListValueNode):
        return flatten_arg(info, arg.values)
    elif isinstance(arg, ValueNode) and hasattr(arg, 'value'):
        return flatten_arg(info, arg.value)
    else:
        # Scalar value
        return arg


def convert_arguments_to_filter_kwargs(info, arguments, filter_arg_name):
    """Converts a list of Argument objects to fields and values pairs to be used in .filter"""
    flattened_args = flatten_args(info, arguments)
    if filter_arg_name not in flattened_args or not flattened_args[filter_arg_name]:
        return None

    args = flattened_args[filter_arg_name]

    filter_kwargs = {}
    for key, input_type in types.FILTER_INPUT_TYPE_MAP.items():
        if key in args:
            for fields in args[key]:
                key = fields['expr']
                value = fields['value']
                filter_kwargs[key] = value

    return filter_kwargs


def convert_arguments_to_order_by_args(info, arguments, order_by_arg_name):
    flattened_args = flatten_args(info, arguments)
    if order_by_arg_name not in flattened_args or not flattened_args[order_by_arg_name]:
        return None

    return flattened_args[order_by_arg_name]


def to_pascal_case(string):
    return string.replace('_', ' ').title().replace(' ', '')



from collections import OrderedDict
from functools import cached_property

import graphene
from django.db.models import Manager
from graphene import String, List
from graphene_django import DjangoObjectType
from graphene_django.types import ErrorType as _ErrorType, DjangoObjectTypeOptions, construct_fields

from graphene_django.filter.utils import get_filtering_args_from_filterset
from graphene_django.utils import get_model_fields

from autographql.filters.types import AuthAutoFilterInputObjectType
from autographql.optimizer import query


class ErrorType(_ErrorType):
    errors = graphene.List(lambda: ErrorType)


class OrderByType(List):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('of_type', String)
        super().__init__(*args, **kwargs)


class AutoDjangoObjectTypeOptions(DjangoObjectTypeOptions):
    input_fields = None

    @cached_property
    def filter_input_type(self):
        return type(self.model.__name__ + 'FilterInput', (AuthAutoFilterInputObjectType,), {
            'Meta': {
                'model': self.model,
                'input_fields': self.input_fields,
            },
        })


class AutoDjangoObjectType(DjangoObjectType):
    """
    Cast select_related to queryset for ForeignKeys of model
    """
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(
            cls,
            model=None,
            registry=None,
            skip_registry=False,
            only_fields=None,  # deprecated in favour of `fields`
            fields=None,
            exclude_fields=None,  # deprecated in favour of `exclude`
            exclude=None,
            filter_fields=None,
            filterset_class=None,
            connection=None,
            connection_class=None,
            use_connection=None,
            interfaces=(),
            convert_choices_to_enum=True,
            _meta=None,
            **options
    ):
        if not _meta:
            _meta = AutoDjangoObjectTypeOptions(cls)

        django_fields = construct_fields(model, registry, fields, exclude, convert_choices_to_enum)
        model_fields = get_model_fields(model)
        # We only will include fields that are included in the model type to avoid exposing anything we shouldn't
        model_fields = filter(lambda m: m[0] in django_fields.keys(), model_fields)
        _meta.input_fields = [f[0] for f in model_fields]

        super().__init_subclass_with_meta__(
            model=model,
            registry=registry,
            skip_registry=skip_registry,
            only_fields=only_fields,  # deprecated in favour of `fields`
            fields=fields,
            exclude_fields=exclude_fields,  # deprecated in favour of `exclude`
            exclude=exclude,
            filter_fields=filter_fields,
            filterset_class=filterset_class,
            connection=connection,
            connection_class=connection_class,
            use_connection=use_connection,
            interfaces=interfaces,
            convert_choices_to_enum=convert_choices_to_enum,
            _meta=_meta,
            **options
        )

    @classmethod
    def get_optimized_queryset(cls, info, args=OrderedDict()):
        queryset = cls._meta.model.objects
        return cls.optimize_queryset(queryset, info, args)

    @classmethod
    def optimize_queryset(cls, queryset, info, args=OrderedDict()):
        queryset = cls.get_filtered_queryset(queryset, info, args)
        queryset = query.QueryOptimizer(info).optimize(queryset)

        return queryset

    @classmethod
    def get_filtered_queryset(cls, queryset, info, args):
        queryset = cls.get_queryset(queryset, info)

        if hasattr(queryset, 'for_user'):
            queryset = queryset.for_user(info.context.user)

        # Apply filterset class if it exists
        filterset_class = cls._meta.filterset_class
        if filterset_class:
            filtering_args = get_filtering_args_from_filterset(filterset_class, cls)
            filter_kwargs = {k: v for k, v in args.items() if k in filtering_args}
            if len(filter_kwargs) > 0:
                queryset = filterset_class(
                    data=filter_kwargs,
                    queryset=queryset,
                    request=info.context,
                ).qs

        # Apply where filters if they exist
        if 'where' in args and args['where']:
            filter_input = args['where']
            user = info.context.user
            lookup = filter_input.get_q_lookup(user)
            if lookup:
                queryset = queryset.filter(lookup)

        # Apply order by if it exists
        if 'orderBy' in args and args['orderBy']:
            order_by_input = args['orderBy']
            queryset = queryset.order_by(*order_by_input)

        if isinstance(queryset, Manager):
            queryset = queryset.all()

        return queryset

    @classmethod
    def get_node(cls, info, id):
        try:
            queryset = cls.get_optimized_queryset(info)
            return queryset.get(pk=id)
        except cls._meta.model.DoesNotExist:
            return None



from graphene_django.registry import get_global_registry


def get_input_type(field):
    registry = get_global_registry()
    related_model = field.related_model
    model_type = registry.get_type_for_model(related_model)
    return model_type._meta.filter_input_type



import graphene


NOT = 'NOT'
AND = 'AND'
OR = 'OR'


class LookupInputField(graphene.InputField):
    def __init__(self, lookup_class, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lookup_class = lookup_class


class LogicalInputField(graphene.List):
    @classmethod
    def get_operator(cls):
        raise NotImplementedError


class LogicalAndInputField(LogicalInputField):
    @classmethod
    def get_operator(cls):
        return AND


class LogicalOrInputField(LogicalInputField):
    @classmethod
    def get_operator(cls):
        return OR


class LogicalNotInputField(LogicalInputField):
    @classmethod
    def get_operator(cls):
        return NOT



import graphene
from django.core.exceptions import PermissionDenied
from graphene import InputField, ClientIDMutation
from graphene.types.utils import yank_fields_from_attrs
from graphene_django.forms.mutation import DjangoModelFormMutation
from graphene_django.rest_framework.mutation import SerializerMutationOptions
from graphql_relay import from_global_id
from graphql_relay.connection.arrayconnection import offset_to_cursor
from rest_framework.exceptions import ErrorDetail

from autographql.auth.constants import PERMISSION_DENIED_MESSAGE
from autographql.auth.utils import get_model_permission, CREATE, DELETE, UPDATE
from autographql.converters import get_input_fields_from_serializer
from autographql.fields import OptimizedField
from autographql.types import ErrorType


class SerializerMutation(ClientIDMutation):
    class Meta:
        abstract = True

    errors = graphene.List(
        ErrorType,
        description='May contain more than one error for same field.'
    )

    @classmethod
    def __init_subclass_with_meta__(
            cls,
            serializer_class=None,
            method=None,
            _meta=None,
            **options
    ):

        if not serializer_class:
            raise Exception('serializer_class is required for the SerializerMutation')

        if method not in ('create', 'update'):
            raise Exception('meta method must be either create or update')

        serializer = serializer_class()
        input_fields = get_input_fields_from_serializer(serializer, method)

        if not _meta:
            _meta = SerializerMutationOptions(cls)
        _meta.serializer_class = serializer_class
        _meta.method = method

        input_fields = yank_fields_from_attrs(
            input_fields,
            _as=InputField,
        )
        super(SerializerMutation, cls).__init_subclass_with_meta__(_meta=_meta, input_fields=input_fields, **options)

    @classmethod
    def mutate_and_get_payload(cls, root, info, **input):
        input = cls.convert_global_id_inputs(cls.Input, **input)
        serializer = cls.get_serializer(root, info, **input)

        if serializer.is_valid():
            return cls.perform_mutate(serializer, info)
        else:
            errors = [
                ErrorType(field=key, messages=value)
                for key, value in serializer.errors.items()
            ]

            return cls(errors=errors)

    @classmethod
    def get_serializer(cls, root, info, **input):
        return cls._meta.serializer_class(data=input)

    @classmethod
    def convert_global_id_inputs(cls, input_class, **input):
        cleaned_input = {}
        for field in input:
            value = input[field]
            input_field = getattr(input_class, field)
            if isinstance(input_field, graphene.GlobalID):
                cleaned_input[field] = int(from_global_id(input[field])[1]) if input[field] else None
            elif isinstance(value, dict):
                next_type = input_field.type
                if hasattr(next_type, 'of_type') and next_type.of_type:
                    next_type = next_type.of_type
                cleaned_input[field] = cls.convert_global_id_inputs(next_type, **value)
            else:
                cleaned_input[field] = input[field]
        return cleaned_input

    @classmethod
    def perform_mutate(cls, serializer, info):
        obj = serializer.save()
        return cls(errors=None, **obj)

    @classmethod
    def get_serializer_errors(cls, serializer):
        errors = [
            cls._get_serializer_errors(key, value)
            for key, value in serializer.errors.items()
        ]

        return errors

    @classmethod
    def _get_serializer_errors(cls, key, value):
        """Recursive helper to build a nested ErrorType object"""
        # Base case, ErrorDetail
        if isinstance(value, list):
            return ErrorType(field=key, messages=value)

        # Recursive call
        if isinstance(value, dict):
            return ErrorType(
                field=key,
                messages=[ErrorDetail('One or more nested fields is not valid.')],
                errors=[
                    cls._get_serializer_errors(key2, value2)
                    for key2, value2 in value.items()
                ]
            )


class CrudSerializerMutationOptions(SerializerMutationOptions):
    type = None
    permission = None


class CrudSerializerMutation(SerializerMutation):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        permission=None,
        type=None,
        _meta=None,
        **options
    ):
        if not type:
            raise Exception('type is required for the CrudSerializerMutation')

        if not _meta:
            _meta = CrudSerializerMutationOptions(cls)

        _meta.type = type
        _meta.permission = permission

        super(CrudSerializerMutation, cls).__init_subclass_with_meta__(_meta=_meta, **options)


class CreateSerializerMutation(CrudSerializerMutation):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, **options):
        super(CreateSerializerMutation, cls).__init_subclass_with_meta__(
            method='create',
            **options
        )
        cls._meta.fields['edge'] = OptimizedField(cls._meta.type._meta.connection.Edge)

    @classmethod
    def mutate_and_get_payload(cls, root, info, **input):
        # Permission check
        if not info.context.user.has_perm(cls._meta.permission):
            raise PermissionDenied(PERMISSION_DENIED_MESSAGE)

        cleaned_input = cls.convert_global_id_inputs(cls.Input, **input)
        return cls.create(root, info, **cleaned_input)

    @classmethod
    def create(cls, root, info, **input):
        serializer_class = cls._meta.serializer_class
        serializer = serializer_class(data=input, context={'request': info.context})
        if serializer.is_valid():
            instance = serializer.save()
        else:
            errors = cls.get_serializer_errors(serializer)
            return cls(errors=errors)

        edge = cls._meta.type._meta.connection.Edge(cursor=offset_to_cursor(0), node=instance)
        kwargs = {}
        kwargs['edge'] = edge

        return cls(errors=None, **kwargs)


class UpdateSerializerMutation(CrudSerializerMutation):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, **options):
        super(UpdateSerializerMutation, cls).__init_subclass_with_meta__(
            method='update',
            **options,
        )
        cls._meta.fields['edge'] = OptimizedField(cls._meta.type._meta.connection.Edge)

    @classmethod
    def mutate_and_get_payload(cls, root, info, **input):
        cleaned_input = cls.convert_global_id_inputs(cls.Input, **input)
        return cls.update(root, info, **cleaned_input)

    @classmethod
    def update(cls, root, info, **input):
        if 'id' not in input:
            return cls(errors=[
                ErrorType(
                    field='id',
                    messages=[ErrorDetail('The id field is required.')],
                )
            ])

        serializer_class = cls._meta.serializer_class
        model_class = serializer_class.Meta.model
        instance = model_class.objects.get(id=input['id'])

        # Permission check
        if not info.context.user.has_perm(cls._meta.permission, instance):
            raise PermissionDenied(PERMISSION_DENIED_MESSAGE)

        serializer = serializer_class(
            instance,
            data=input,
            context={'request': info.context},
            partial=True
        )
        if serializer.is_valid():
            instance = serializer.save()
        else:
            errors = cls.get_serializer_errors(serializer)
            return cls(errors=errors)

        if getattr(instance, '_prefetched_objects_cache', None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        kwargs = {}
        edge = cls._meta.type._meta.connection.Edge(cursor=offset_to_cursor(0), node=instance)
        kwargs['edge'] = edge

        return cls(errors=None, **kwargs)


class DjangoSerializerMutationFieldFactory(object):
    class Meta:
        type = None
        serializer_class = None

    @classmethod
    def get_create_field(cls):
        return cls.get_create_mutation().Field()

    @classmethod
    def get_create_mutation(cls):
        # Get required permissions
        add_permission = get_model_permission(cls.Meta.type._meta.model, CREATE)

        class CreateInstance(CreateSerializerMutation):
            class Meta:
                name = 'Create' + cls.Meta.serializer_class.Meta.model.__name__ + 'Payload'
                serializer_class = cls.Meta.serializer_class
                type = cls.Meta.type
                permission = add_permission

        return CreateInstance

    @classmethod
    def get_update_field(cls):
        return cls.get_update_mutation().Field()

    @classmethod
    def get_update_mutation(cls):
        # Get required permissions
        change_permission = get_model_permission(cls.Meta.type._meta.model, UPDATE)

        class UpdateInstance(UpdateSerializerMutation):
            class Meta:
                name = 'Update' + cls.Meta.serializer_class.Meta.model.__name__ + 'Payload'
                serializer_class = cls.Meta.serializer_class
                type = cls.Meta.type
                permission = change_permission

        return UpdateInstance

    @classmethod
    def get_delete_field(cls):
        return cls.get_delete_mutation().Field()

    @classmethod
    def get_delete_mutation(cls):
        # Get required permissions
        delete_permission = get_model_permission(cls.Meta.type._meta.model, DELETE)

        class DeleteInstance(ClientIDMutation):
            class Meta:
                name = 'Delete' + cls.Meta.serializer_class.Meta.model.__name__ + 'Payload'

            class Input:
                id = graphene.String()

            deletedId = graphene.String()
            count = graphene.Int()

            @classmethod
            def mutate_and_get_payload(inner_cls, root, info, **input):
                return inner_cls.delete(root, info, **input)

            @classmethod
            def delete(inner_cls, root, info, **input):
                deletedId = from_global_id(input['id'])[1]
                model_class = cls.Meta.serializer_class.Meta.model
                instance = model_class.objects.get(id=deletedId)

                # Permission check
                if not info.context.user.has_perm(delete_permission, instance):
                    raise PermissionDenied(PERMISSION_DENIED_MESSAGE)

                deleted = instance.delete()

                return DeleteInstance(
                    deletedId=input['id'],
                    count=deleted[0]
                )

        return DeleteInstance


class CoreDjangoModelFormMutation(DjangoModelFormMutation):
    class Meta:
        abstract = True

    errors = graphene.List(ErrorType)



import functools

from django.db.models import ManyToOneRel, Prefetch
from graphene import GlobalID
from graphene.types.mutation import MutationOptions
from graphene.types.resolver import attr_resolver, dict_or_attr_resolver, dict_resolver
from graphene_django import DjangoObjectType
from graphene_django_optimizer.query import QueryOptimizer as _QueryOptimizer
from graphene_django_optimizer.query import QueryOptimizerStore as _QueryOptimizerStore
from graphql.execution.values import get_argument_values
from graphql.language.ast import (
    FragmentSpreadNode,
    InlineFragmentNode,
)

from autographql.auth.query import AuthQueryOptimizer
from autographql.optimizer.utils import remove_prefix, combine_querysets


class QueryOptimizer(_QueryOptimizer):
    def __init__(self, info):
        super(QueryOptimizer, self).__init__(info)
        self.auth_optimizer = AuthQueryOptimizer(info)
        self.variable_values = info.variable_values

    def optimize(self, queryset):
        queryset = super().optimize(queryset)
        queryset = self.auth_optimizer.optimize(queryset)
        queryset = combine_querysets([queryset])
        return queryset

    def _optimize_gql_selections(self, field_type, field_ast):
        store = QueryOptimizerStore()
        store.abort_only_optimization()

        # Optimize queryset
        selection_set = field_ast.selection_set
        if not selection_set:
            return store
        optimized_fields_by_model = {}
        schema = self.root_info.schema
        graphql_type = schema.get_type(field_type.name)
        possible_types = self._get_possible_types(graphql_type)
        for selection in selection_set.selections:
            if isinstance(selection, InlineFragmentNode):
                self.handle_inline_fragment(
                    selection, schema, possible_types, store)
            else:
                name = selection.name.value
                if isinstance(selection, FragmentSpreadNode):
                    self.handle_fragment_spread(store, name, field_type)
                else:
                    for possible_type in possible_types:
                        selection_field_def = possible_type.fields.get(name)
                        if not selection_field_def:
                            continue

                        graphene_type = possible_type.graphene_type
                        # Check if graphene type is a relay connection or a relay edge or mutation payload
                        if hasattr(graphene_type._meta, 'node') or (
                            hasattr(graphene_type, 'cursor') and
                            hasattr(graphene_type, 'node')
                        ) or (
                            isinstance(graphene_type._meta, MutationOptions)
                        ):
                            relay_store = self._optimize_gql_selections(
                                self._get_type(selection_field_def),
                                selection,
                            )
                            store.append(relay_store)
                            try:
                                from django.db.models import DEFERRED  # noqa: F401
                            except ImportError:
                                store.abort_only_optimization()
                        else:
                            model = getattr(graphene_type._meta, 'model', None)
                            # if model and name not in optimized_fields_by_model:
                            # Always optimize in case of duplicate field names
                            if model:
                                # Its possible that model is None like in the case of processing the pageInfo field
                                field_model = optimized_fields_by_model[name] = model
                                if field_model == model:
                                    self._optimize_field(
                                        store,
                                        model,
                                        selection,
                                        selection_field_def,
                                        possible_type,
                                    )
        return store

    def _get_name_from_resolver(self, resolver):
        resolver_fn = resolver
        optimization_hints = self._get_optimization_hints(resolver)
        if optimization_hints:
            name = optimization_hints.model_field
            if name:
                return name

        while isinstance(resolver_fn, functools.partial):
            if hasattr(resolver_fn, 'func') and (
                    resolver_fn.func == attr_resolver or
                    resolver_fn.func == dict_resolver or
                    resolver_fn.func == dict_or_attr_resolver
            ):
                return resolver_fn.args[0]

            resolver_fn = resolver_fn.args[0]

        if self._is_resolver_for_id_field(resolver):
            return 'id'

        # Unknown resolver type, just extract the field name from the resolver name
        return remove_prefix(resolver_fn.__name__, 'resolve_')

    def _optimize_field_by_name(self, store, model, selection, field_def):
        name = self._get_name_from_resolver(field_def.resolve)
        if not name:
            return False
        model_field = self._get_model_field_from_name(model, name)
        if not model_field:
            return False
        if self._is_foreign_key_id(model_field, name):
            store.only(name)
            return True
        if (
                model_field.many_to_one or
                model_field.one_to_one or
                model_field.one_to_many or
                model_field.many_to_many
        ):
            node_type = self._get_type(field_def)
            field_store = self._optimize_gql_selections(
                node_type,
                selection,
                # parent_type,
            )

            if isinstance(model_field, ManyToOneRel):
                field_store.only(model_field.field.name)

            arguments = get_argument_values(field_def, selection, self.variable_values)
            related_queryset = self.get_queryset(node_type, arguments)
            store.prefetch_related(name, field_store, related_queryset)
            return True
        if not model_field.is_relation:
            store.only(name)
            return True
        return False

    def get_queryset(self, node_type, args):
        gtype = node_type.graphene_type

        if hasattr(gtype._meta, 'node'):
            gtype = gtype._meta.node

        model = gtype._meta.model
        queryset = gtype.get_filtered_queryset(model.objects, self.root_info, args)
        return self.auth_optimizer.optimize(queryset)

    def _is_resolver_for_id_field(self, resolver):
        resolve_id = DjangoObjectType.resolve_id
        # For python 2 unbound method:
        if hasattr(resolve_id, 'im_func'):
            resolve_id = resolve_id.im_func

        # Check to see if its a relay GlobalID
        resolver_fn = resolver
        if isinstance(resolver_fn, functools.partial):
            if resolver_fn.func == GlobalID.id_resolver:
                resolver_fn = resolver_fn.args[0]

        return resolver_fn == resolve_id


class QueryOptimizerStore(_QueryOptimizerStore):
    def optimize_queryset(self, queryset):
        if len(self.select_list) > 0:
            # An empty select_list will have the queryset select everything possible
            queryset = queryset.select_related(*self.select_list)
        if len(self.prefetch_list) > 0:
            queryset = queryset.prefetch_related(*self.prefetch_list)

        if self.only_list:
            queryset = queryset.only(*self.only_list)
        return queryset

    def prefetch_related(self, name, store, queryset):
        """Overridden prefetch_related always use the prefetch object"""
        queryset = store.optimize_queryset(queryset)
        self.prefetch_list.append(Prefetch(name, queryset=queryset))



import six

from functools import wraps
from django.core.exceptions import PermissionDenied


def context(f):
    def decorator(func):
        def wrapper(*args, **kwargs):
            info = args[f.__code__.co_varnames.index('info')]
            return func(info.context, *args, **kwargs)
        return wrapper
    return decorator


def user_passes_test(test_func, error_message=None):
    def decorator(f):
        @wraps(f)
        @context(f)
        def wrapper(context, *args, **kwargs):
            if test_func(context.user):
                return f(*args, **kwargs)
            raise PermissionDenied(error_message)
        return wrapper
    return decorator


def permission_required(perm, obj=None):
    def check_perms(user):
        if isinstance(perm, six.string_types):
            perms = (perm,)
        else:
            perms = perm

        if user.has_perms(perms, obj):
            return True
        return False
    return user_passes_test(check_perms, 'You do not have permission to perform this action')


login_required = user_passes_test(lambda u: u.is_authenticated, 'User must be authenticated')
